"""Provides example models and tutorials for the `amisc` package.

Includes:

- `models`: Example models for testing and demonstration.
- `tutorial`: Example tutorials for using the `amisc` package.
"""